import pygame as pg
from pg_extended.Types import Background
from pg_extended.Util import Misc
from pg_extended.Core import DynamicValue, CallbackSet
from pg_extended.UI.Elements.Section import Section
from pg_extended.UI.Elements.Circle import Circle

class Slider():
  def __init__(self, orientation: str, section: Section, dragElement: Section | Circle, valueRange: list[float] | tuple[float], scrollSpeed: float, filledSliderBackground: Background, callback: CallbackSet = None, hoverToScroll: bool | None = True):
    self.orientation = orientation
    self.section = section
    self.valueRange = valueRange
    self.scrollSpeed = scrollSpeed
    self.dragElement = dragElement
    self.filledSliderBackground = filledSliderBackground
    self.callback = callback
    self.hoverToScroll = hoverToScroll

    self.value = self.valueRange[0]
    self.pressed = False
    self.active = True
    self.activeDraw = True
    self.activeUpdate = True
    self.activeEvents = True
    self.lazyUpdate = True
    self.lazyUpdateOverride = False

    if not self.orientation in ('vertical', 'horizontal'):
      raise ValueError('Slider orientation must be \'vertical\' or \'horizontal\'')

    if isinstance(self.dragElement, Section):
      self.dragElementType = 'section'
    else:
      self.dragElementType = 'circle'

    def getDragElementPos(axis: str, elementType: str, slider: Slider):
      returnValue = None
      axis = axis
      elementType = elementType
      slider = slider
      section = slider.section
      sliderValue = slider.value
      sliderValueRange = slider.valueRange
      dragElement = slider.dragElement

      if axis == 'x':
        if elementType == 'section':
          valueMappingCoords = (section.x, section.x + section.width - dragElement.width)
        else:
          valueMappingCoords = (section.x + dragElement.radius, section.x + section.width - dragElement.radius)

        returnValue = Misc.mapRange(sliderValue, sliderValueRange[0], sliderValueRange[1], valueMappingCoords[0], valueMappingCoords[1])
      else:
        if elementType == 'section':
          valueMappingCoords = (section.y, section.y + section.height - dragElement.height)
        else:
          valueMappingCoords = (section.y + dragElement.radius, section.y + section.height - dragElement.radius)

        returnValue = Misc.mapRange(sliderValue, sliderValueRange[0], sliderValueRange[1], valueMappingCoords[0], valueMappingCoords[1])

      if returnValue < valueMappingCoords[0]: return valueMappingCoords[0]
      elif returnValue > valueMappingCoords[1]: return valueMappingCoords[1]
      else: return returnValue

    if self.dragElementType == 'circle':
      if self.orientation == 'horizontal':
        self.dragElement.dimensions['x'] = DynamicValue(
          getDragElementPos,
          args={'axis': 'x', 'elementType': 'circle', 'slider': self}
        )

        self.dragElement.dimensions['y'] = DynamicValue(
          lambda section: section.y + (section.height / 2),
          args={'section': self.section}
        )

      else:
        self.dragElement.dimensions['x'] = DynamicValue(
          lambda section: section.x + (section.width / 2),
          args={'section': self.section}
        )

        self.dragElement.dimensions['y'] = DynamicValue(
          getDragElementPos,
          args={'axis': 'y', 'elementType': 'circle', 'slider': self}
        )

    else:
      if self.orientation == 'horizontal':
        self.dragElement.dimensions['x'] = DynamicValue(
          getDragElementPos,
          args={'axis': 'x', 'elementType': 'section', 'slider': self}
        )

        self.dragElement.dimensions['y'] = DynamicValue(
          lambda section, dragElement: section.y + ((section.height - dragElement.height) / 2),
          args={'section': self.section, 'dragElement': self.dragElement}
        )

      else:
        self.dragElement.dimensions['x'] = DynamicValue(
          lambda section, dragElement: section.x + ((section.width - dragElement.width) / 2),
          args={'section': self.section, 'dragElement': self.dragElement}
        )

        self.dragElement.dimensions['y'] = DynamicValue(
          getDragElementPos,
          args={'axis': 'y', 'elementType': 'section', 'slider': self}
        )

    if self.dragElementType == 'section':
      if self.orientation == 'horizontal':
        self.mapPosition = DynamicValue(
          lambda element: element.x + (element.width / 2),
          args={'element': self.dragElement}
        )

      else:
        self.mapPosition = DynamicValue(
          lambda element: element.y + (element.height / 2),
          args={'element': self.dragElement}
        )

    else:
      if self.orientation == 'horizontal':
        self.mapPosition = self.dragElement.dimensions['x']
      else:
        self.mapPosition = self.dragElement.dimensions['y']

    filledSliderWidth = None
    filledSliderHeight = None

    if self.orientation == 'horizontal':
      filledSliderWidth = DynamicValue(
        lambda mapPosition, section: max(0, mapPosition.value - section.x),
        args={'mapPosition': self.mapPosition, 'section': self.section}
      )

      filledSliderHeight = self.section.dimensions['height']
    else:
      filledSliderWidth = self.section.dimensions['width']

      filledSliderHeight = DynamicValue(
        lambda mapPosition, section: max(0, mapPosition.value - section.y),
        args={'mapPosition': self.mapPosition, 'section': self.section}
      )

    self.filledSlider = Section(
      {
        'x': self.section.dimensions['x'],
        'y': self.section.dimensions['y'],
        'width': filledSliderWidth,
        'height': filledSliderHeight
      }, self.filledSliderBackground, self.section.borderRadius
    )

  def update(self):
    if not (self.active and self.activeUpdate):
      return None

    self.section.update()
    self.dragElement.update()
    self.mapPosition.resolveValue()
    self.filledSlider.update()

  def updateValue(self):
    if not self.active:
      return None

    mousePos = pg.mouse.get_pos()

    if self.orientation == 'horizontal':
      relativePos = mousePos[0]
      start = self.section.x
      end = self.section.x + self.section.width
    else:
      relativePos = mousePos[1]
      start = self.section.y
      end = self.section.y + self.section.height

    if relativePos < start:
      relativePos = start
    elif relativePos > end:
      relativePos = end

    self.value = Misc.mapRange(relativePos, start, end, self.valueRange[0], self.valueRange[1])

    self.dragElement.update()
    self.mapPosition.resolveValue()
    self.filledSlider.update()

  def draw(self, surface: pg.Surface):
    if not (self.active and self.activeDraw):
      return None

    self.section.draw(surface)
    self.filledSlider.draw(surface)
    self.dragElement.draw(surface)

  def handleCallback(self, trigger):
    if self.callback is not None:
      self.callback.call(trigger, {'value': self.value})

  def checkEvent(self, event: pg.Event) -> bool:
    if not (self.active and self.activeEvents):
      return None

    if event.type == pg.MOUSEBUTTONDOWN and event.button == 1 and self.section.rect.collidepoint(pg.mouse.get_pos()):
      self.pressed = True

      self.updateValue()

      self.handleCallback('mouseDown')

    if event.type == pg.MOUSEBUTTONUP and event.button == 1:
      if self.pressed:
        self.pressed = False

        self.handleCallback('mouseUp')

        return True

      return False

    if self.pressed and event.type == pg.MOUSEMOTION:
      self.updateValue()

      self.handleCallback('mouseDrag')

      return True

    if event.type == pg.MOUSEWHEEL:
      self.pressed = False
      scroll = False
      updatedValue = self.value

      if self.hoverToScroll:
        if self.section.rect.collidepoint(pg.mouse.get_pos()):
          scroll = True
      else:
        scroll = True

      if scroll:
        if event.x > 0 or event.y > 0:
          updatedValue += self.scrollSpeed
        elif event.x < 0 or event.y < 0:
          updatedValue -= self.scrollSpeed

        if updatedValue < min(self.valueRange[0], self.valueRange[1]):
          updatedValue = min(self.valueRange[0], self.valueRange[1])
        elif updatedValue > max(self.valueRange[0], self.valueRange[1]):
          updatedValue = max(self.valueRange[0], self.valueRange[1])

        if self.value != updatedValue:
          self.value = updatedValue

          self.update()

          self.handleCallback('scroll')

          return True

        return False

      return False

    return False



import pygame as pg
from pg_extended.Types import TileIdentifier

class SpriteAnimation:
  def __init__(self, animation: list[tuple[str, TileIdentifier]] | tuple[tuple[str, TileIdentifier]]):
    self.animationRaw = animation

    self.scene: 'Scene' = None # type: ignore
    self.sprites: list[pg.Surface] = []

  def initiate(self, scene: 'Scene'): # type: ignore
    self.scene = scene

    for tileDetails in self.animationRaw:
      atlasID, tileID = tileDetails

      sprite = self.scene.textureAtlases[atlasID].getTile(tileID)

      if sprite is None: continue

      self.sprites.append(sprite)



import os
import pygame as pg
from pg_extended.Core import DynamicValue
from pg_extended.UI.Elements.Section import Section

ALIGNMENT_MAP = {
  'left-top': 'topleft',
  'left-center': 'midleft',
  'left-bottom': 'bottomleft',
  'center-top': 'midtop',
  'center-center': 'center',
  'center-bottom': 'midbottom',
  'right-top': 'topright',
  'right-center': 'midright',
  'right-bottom': 'bottomright',
}

class TextBox:
  def __init__(self, section: Section, text: str, fontPath: str, textColor: pg.Color, fontSize: DynamicValue = None):
    self.section = section
    self.text = text
    self.fontPath = fontPath
    self.fontSize = fontSize
    self.textColor = textColor

    self.alignTextHorizontal = 'center'
    self.alignTextVertical = 'center'
    self.drawSectionDefault = False
    self.paddingLeft = 0
    self.paddingRight = 0
    self.paddingLeftStr = None
    self.paddingRightStr = None
    self.active = True
    self.activeDraw = True
    self.activeUpdate = True
    self.lazyUpdate = True
    self.lazyUpdateOverride = False
    self.textSurface: pg.Surface = None
    self.textRect: pg.Rect = None

    self.update()

  def update(self):
    if not (self.active and self.activeUpdate):
      return None

    self.section.update()

    if self.text == '':
      return None

    if self.fontSize:
      self.fontSize.resolveValue()
      fontSize = self.fontSize.value
    else:
      fontSize = .6 * self.section.height

    if os.path.exists(self.fontPath):
      self.font = pg.font.Font(self.fontPath, int(fontSize))
    else:
      matched = pg.font.match_font(self.fontPath)
      if matched:
        self.font = pg.font.Font(matched, int(fontSize))
      else:
        self.font = pg.font.Font(None, int(fontSize))

    self.paddingLeftStr = ' ' * self.paddingLeft
    self.paddingRightStr = ' ' * self.paddingRight
    self.textSurface = self.font.render(f'{self.paddingLeftStr}{self.text}{self.paddingRightStr}', True, self.textColor)

    if self.textColor.a < 255:
      self.textSurface.set_alpha(self.textColor.a)

    key = f'{self.alignTextHorizontal}-{self.alignTextVertical}'
    pos_attr = ALIGNMENT_MAP[key]

    self.textRect = self.textSurface.get_rect(**{pos_attr: getattr(self.section.rect, pos_attr)})

  def draw(self, surface: pg.Surface, drawSection: bool = None):
    if not (self.active and self.activeDraw):
      return None

    if (drawSection is None and self.drawSectionDefault) or drawSection:
      self.section.draw(surface)

    if self.text == '':
      return None

    surface.blit(self.textSurface, self.textRect)



import pygame as pg
from pg_extended.Core import DynamicValue
from pg_extended.Game.Scene import Scene

class ViewPort:
  def __init__(self, x: DynamicValue, y: DynamicValue, scale: float):
    self.x = x
    self.y = y
    self.scale = scale

    self.scaledLevelSurface: pg.Surface = None
    self.preRendererdView: pg.Surface = None
    self.parentSurface: pg.Surface = None
    self.scalingMultiplier: float = 10.0
    self.scalingFactor: float = 0.0
    self.scaledTileWidth: float = 0.0
    self.scaledTileHeight: float = 0.0
    self.scenePosition: tuple[float, float] = (0.0, 0.0)
    self.locked: bool = True
    self.lazyRender: bool = False

  def initiate(self, surface: pg.Surface, scene: Scene):
    self.parentSurface = surface
    self.scene = scene

    self.locked = False
    self.update()

  def update(self):
    self.scalingFactor = self.parentSurface.get_height() / (self.scale * self.scalingMultiplier * self.scene.activeLevel.tileHeight)

    self.scaledTileWidth = self.scene.activeLevel.tileWidth * self.scalingFactor
    self.scaledTileHeight = self.scene.activeLevel.tileHeight * self.scalingFactor

    self.x.resolveValue()
    self.y.resolveValue()

    self.scenePosition = (-self.scaledTileWidth * self.x.value, -self.scaledTileHeight * self.y.value)

  def renderScene(self):
    if self.locked: return None

    self.scaledLevelSurface = pg.transform.scale_by(self.scene.activeLevel.surface, self.scalingFactor)

    self.preRenderedView = pg.Surface(self.parentSurface.get_size(), pg.SRCALPHA)

    self.preRenderedView.blit(self.scaledLevelSurface)

    for entity in self.scene.activeLevel.entities:
      if entity.sprite is None: continue

      entitySprite = pg.transform.scale(entity.sprite, (int(entity.widthPX * self.scalingFactor), int(entity.heightPX * self.scalingFactor)))

      entityPos = (self.scaledTileWidth * entity.x, self.scaledTileHeight * entity.y)

      self.preRenderedView.blit(entitySprite, entityPos)

  def draw(self):
    if self.locked: return None

    self.parentSurface.blit(self.preRenderedView, self.scenePosition)



from __future__ import annotations
from typing import Any
from pg_extended.Types import CallableLike
import types

class DynamicValue:
  def __init__(self, ref: Any, lookup: str | None = None, args: dict[str, Any] | None = None, percent: int | float | None = None):
    self.reference = ref
    self.lookup = lookup
    self.args = args
    self.percent = percent
    self.value: Any = None
    self.resolveValue: CallableLike = None

    self.assignResolveMethod()

    self.resolveValue()

  def _IFPer(self):
    self.value = self.reference / 100 * self.percent

  def _dictLookup(self):
    self.value = self.reference[self.lookup]

  def _dictLookupPer(self):
    self.value = self.reference[self.lookup] / 100 * self.percent

  def _CV(self):
    self.reference.resolveValue()
    self.value = self.reference.value

  def _CVPer(self):
    self.reference.resolveValue()
    self.value = self.reference.value / 100 * self.percent

  def _call(self):
    self.value = self.reference()

  def _callPer(self):
    self.value = self.reference() / 100 * self.percent

  def _callArgs(self):
    self.value = self.reference(**self.args)

  def _callArgsPer(self):
    self.value = self.reference(**self.args) / 100 * self.percent

  def _objLookup(self):
    self.value = getattr(self.reference, self.lookup)

  def _objLookupPer(self):
    self.value = getattr(self.reference, self.lookup) / 100 * self.percent

  def _direct(self):
    self.value = self.reference

  def assignResolveMethod(self):
    from pg_extended.Core.Base.AnimatedValue import AnimatedValue

    # numbers with a percent value given
    if isinstance(self.reference, (int, float)) and self.percent is not None:
      self.resolveValue = self._IFPer

    # dicts
    elif isinstance(self.reference, dict) and self.lookup is not None:
      if self.percent is None:
        self.resolveValue = self._dictLookup
      else:
        self.resolveValue = self._dictLookupPer

    # DV or AV
    elif isinstance(self.reference, DynamicValue) or isinstance(self.reference, AnimatedValue):
      if self.percent is None:
        self.resolveValue = self._CV
      else:
        self.resolveValue = self._CVPer

    # callable
    elif isinstance(self.reference, (types.FunctionType | types.BuiltinFunctionType | types.MethodType)):
      if self.args is None:
        if self.percent is None:
          self.resolveValue = self._call
        else:
          self.resolveValue = self._callPer
      else:
        if self.percent is None:
          self.resolveValue = self._callArgs
        else:
          self.resolveValue = self._callArgsPer

    # none of the above and lookup is provided, assume it's a class + attribute
    elif isinstance(self.lookup, str):
      if self.percent is None:
        self.resolveValue = self._objLookup
      else:
        self.resolveValue = self._objLookupPer

    # dump everything else into direct
    else:
      self.resolveValue = self._direct



from pg_extended.UI.Elements.Section import Section
from pg_extended.UI.Elements.Circle import Circle
from pg_extended.UI.Elements.TextBox import TextBox
from pg_extended.UI.Elements.Button import Button
from pg_extended.UI.Elements.Toggle import Toggle
from pg_extended.UI.Elements.Slider import Slider
from pg_extended.UI.Elements.TextInput import TextInput

type UIElement = Section | Circle | TextBox | Button | Toggle | Slider | TextInput

__all__ = [
  'Section',
  'Circle',
  'TextBox',
  'Button',
  'Toggle',
  'Slider',
  'TextInput',
  'UIElement'
]



from pg_extended import Types
from pg_extended import Core
from pg_extended import UI
from pg_extended import Game
from pg_extended import Util
from pg_extended import Window

from pg_extended.Types import *
from pg_extended.Core import *
from pg_extended.UI import *
from pg_extended.Game import *
from pg_extended.Util import *
from pg_extended.Window import *

import pygame
pygame.init()



import pygame as pg

CURSOR_CONSTANTS = {
  None: pg.SYSTEM_CURSOR_ARROW,
  'arrow': pg.SYSTEM_CURSOR_ARROW,
  'hand': pg.SYSTEM_CURSOR_HAND,
  'ibeam': pg.SYSTEM_CURSOR_IBEAM
}

class EventManager:
  @staticmethod
  def setCursor(cursor):
    pg.mouse.set_cursor(cursor)

  def handleEvents(self):
    if not self.running:
      return None

    for event in pg.event.get():
      if self.customEventHandler is not None:
        self.customEventHandler(event)

      if event.type == pg.QUIT:
        self.running = False
      else:
        cursorChange = None

        if self.activeScene is not None:
          cursorChange = self.activeScene.handleEvents(event)

        for systemID in self.systemZ:
          if systemID in self.activeSystems:
            if cursorChange is None:
              cursorChange = self.activeSystems[systemID].handleEvents(event)
            else:
              self.activeSystems[systemID].handleEvents(event)

        self.setCursor(CURSOR_CONSTANTS[cursorChange])



import pygame as pg
import pg_extended as pgx

class SceneManager:
  def addScene(self, scene: pgx.Game.Scene, sceneID: str) -> bool:
    if sceneID in self.scenes:
      print(f'A scene with ID: {sceneID} already exists. please enter a unique ID')
      return False

    if not scene.locked:
      print('Provided scene is not locked, switching the scene to locked state and removing its surface.')
      scene.locked = True
      scene.surface = None

    self.scenes[sceneID] = scene

    return True

  def setActiveScene(self, sceneID: str) -> bool:
    interrupted = False

    if not self.scenes[sceneID]:
      print(f'A scene with ID: {sceneID} does not exist. Please provide a valid scene ID')
      interrupted = True
    elif self.scenes[sceneID] == self.activeScene:
      print(f'The scene with ID: {sceneID} is already active.')
      interrupted = True
    else:
      self.activeScene = self.scenes[sceneID]

      if self.running:
        self.initiateActiveScene(self.screen)
        self.resetUI()

    return not interrupted

  def deactivateScene(self) -> bool:
    if self.activeScene is None:
      print('No scene is currently active.')
      return False

    self.activeScene = None

    if self.running:
      self.resetUI()

    return True

  def initiateActiveScene(self, surface: pg.Surface):
    if self.activeScene is not None and self.activeScene.locked:
      self.activeScene.initiate(surface)




import pygame as pg
import pg_extended as pgx

class SystemManager:
  def addSystem(self, system: pgx.UI.System, systemID: str) -> bool:
    if systemID in self.systems:
      print(f'A system with ID: {systemID} already exists. please enter a unique ID')
      return False

    if not system.locked:
      print('Provided system is not locked, switching the system to locked state and removing its surface.')
      system.locked = True
      system.surface = None

    self.systems[systemID] = system

    return True

  def activateSystems(self, systemIDs: list[str] | tuple[str] | str) -> bool:
    interrupted = False

    if not isinstance(systemIDs, str):
      for systemID in systemIDs:
        if not systemID in self.systems:
          print(f'A system with ID: {systemID} does not exist. Automatically skipped this task.')
          interrupted = True
        elif systemID in self.activeSystems:
          print(f'The system with ID: {systemID} is already active.')
          interrupted = True
        else:
          self.activeSystems[systemID] = self.systems[systemID]
    else:
      if not systemIDs in self.systems:
        print(f'A system with ID: {systemIDs} does not exist. Please provide a valid system ID')
        interrupted = True
      elif systemIDs in self.activeSystems:
        print(f'The system with ID: {systemIDs} is already active.')
        interrupted = True
      else:
        self.activeSystems[systemIDs] = self.systems[systemIDs]

    if self.running:
      self.initiateActiveSystems(self.screen)
      self.resetUI()

    return not interrupted

  def deactivateSystems(self, systemIDs: list[str] | tuple[str] | str) -> bool:
    interrupted = False

    if not isinstance(systemIDs, str):
      deleteSystems = []

      for systemID in systemIDs:
        if not systemID in self.activeSystems:
          print(f'A system with ID: {systemID} does not exist or is already deactivated. Automatically skipped this task.')
          interrupted = True
        else:
          self.activeSystems[systemID].locked = True
          del self.activeSystems[systemID].surface
          deleteSystems.append(systemID)

      for systemID in deleteSystems:
        del self.activeSystems[systemID]
    else:
      if systemIDs == 'all':
        for systemID in self.activeSystems:
          self.activeSystems[systemID].locked = True
          del self.activeSystems[systemID].surface

        self.activeSystems = {}
      elif not systemIDs in self.systems:
        print(f'A system with ID: {systemIDs} does not exist or is already deactivated. Automatically skipped this task.')
        interrupted = True
      else:
        self.activeSystems[systemIDs].locked = True
        del self.activeSystems[systemIDs].surface
        del self.activeSystems[systemIDs]

    return not interrupted

  def setSystemZ(self, systemID: str, zIndex: int) -> bool | None:
    keysToRemove = [key for key, value in self.systemZ.items() if value == zIndex]

    for key in keysToRemove:
      del self.systemZ[key]

    self.systemZ[systemID] = zIndex

    self.systemZ = dict(sorted(self.systemZ.items(), key=lambda item: item[1]))

  def initiateActiveSystems(self, surface: pg.Surface):
    for systemID in self.activeSystems:
      if self.activeSystems[systemID].locked:
        self.activeSystems[systemID].initiate(surface)



import pygame as pg
from pg_extended.Util import Misc
from pg_extended.UI.Elements import *

class System:
  def __init__(self, surface: pg.Surface = None, preLoadState: bool = False):
    self.locked = preLoadState

    if not self.locked:
      if not surface:
        self.locked = True
        print('No surface provided, the system is locked by default.\nIt can be initiated manually by providing a surface')
      else:
        self.surface = surface

    self.elements: dict[str, UIElement] = {}
    self.sections: dict[str, Section] = {}
    self.circles: dict[str, Circle] = {}
    self.textBoxes: dict[str, TextBox] = {}
    self.buttons: dict[str, Button] = {}
    self.toggles: dict[str, Toggle] = {}
    self.sliders: dict[str, Slider] = {}
    self.textInputs: dict[str, TextInput] = {}

    self.firstDraw = True

  def addElement(self, element: UIElement, elementID: str):
    if elementID in self.elements:
      raise ValueError(f'An element with id: {elementID} already exists, please enter a unique id.')

    self.elements[elementID] = element

    if isinstance(element, Section):
      self.sections[elementID] = element
    elif isinstance(element, Circle):
      self.circles[elementID] = element
    elif isinstance(element, TextBox):
      self.textBoxes[elementID] = element
    elif isinstance(element, Button):
      self.buttons[elementID] = element
    elif isinstance(element, Toggle):
      self.toggles[elementID] = element
    elif isinstance(element, Slider):
      self.sliders[elementID] = element
    elif isinstance(element, TextInput):
      self.textInputs[elementID] = element

  def addElements(self, elements: dict[str, UIElement]):
    for elementID in elements:
      self.addElement(elements[elementID], elementID)

  def removeElement(self, elementID: str) -> bool:
    if not elementID in self.elements:
      raise ValueError(f'An element with id: {elementID} does not exist, please enter a valid id.')

    element = self.elements[elementID]

    if isinstance(element, Section):
      del self.sections[elementID]
    elif isinstance(element, Circle):
      del self.circles[elementID]
    elif isinstance(element, TextBox):
      del self.textBoxes[elementID]
    elif isinstance(element, Button):
      del self.buttons[elementID]
    elif isinstance(element, Toggle):
      del self.toggles[elementID]
    elif isinstance(element, Slider):
      del self.sliders[elementID]
    elif isinstance(element, TextInput):
      del self.textInputs[elementID]

    del self.elements[elementID]

    return True

  def __validateIDs(self, elementIDs: list[str] | tuple[str] = None) -> list | dict | None:
    if elementIDs == None:
      return self.elements

    if not Misc.allIn(elementIDs, self.elements):
      print('The given iterable contains id(s) that do not exist in this system, please enter a valid iterable')
      return []

    return elementIDs

  def draw(self, elementIDs: list[str] | tuple[str] = None):
    if self.locked:
      print('System is currently locked')
      return None

    idList = self.__validateIDs(elementIDs)

    for elementID in idList:
      if self.elements[elementID].active and self.elements[elementID].activeDraw:
        self.elements[elementID].draw(self.surface)

    self.firstDraw = False

  def update(self, elementIDs: list[str] | tuple[str] = None):
    if self.locked:
      print('System is currently locked')
      return None

    idList = self.__validateIDs(elementIDs)

    for elementID in idList:
      element = self.elements[elementID]

      if element.lazyUpdateOverride or ((not element.lazyUpdate) and element.active):
        element.update()

  def lazyUpdate(self, elementIDs: list[str] | tuple[str] = None):
    if self.locked:
      print('System is currently locked')
      return None

    idList = self.__validateIDs(elementIDs)

    for elementID in idList:
      if self.elements[elementID].active:
        self.elements[elementID].update()

  def handleEvents(self, event: pg.Event) -> str | None:
    if self.locked:
      print('System is currently locked')
      return None

    mousePos = pg.mouse.get_pos()

    changeCursor = None

    for buttonID in self.buttons:
      if self.buttons[buttonID].active:
        self.buttons[buttonID].checkEvent(event)

        if not changeCursor and self.buttons[buttonID].activeEvents:
          if self.buttons[buttonID].section.rect.collidepoint(mousePos):
            changeCursor = 'hand'

    for toggleID in self.toggles:
      if self.toggles[toggleID].active:
        self.toggles[toggleID].checkEvent(event)

        if not changeCursor and self.toggles[toggleID].activeEvents:
          if self.toggles[toggleID].section.rect.collidepoint(mousePos):
            changeCursor = 'hand'

    for sliderID in self.sliders:
      if self.sliders[sliderID].active:
        self.sliders[sliderID].checkEvent(event)

        if not changeCursor and self.sliders[sliderID].activeEvents:
          if self.sliders[sliderID].section.rect.collidepoint(mousePos):
            changeCursor = 'hand'

    for textInputID in self.textInputs:
      if self.textInputs[textInputID].active:
        self.textInputs[textInputID].checkEvent(event)

        if not changeCursor and self.textInputs[textInputID].activeEvents:
          if self.textInputs[textInputID].section.rect.collidepoint(mousePos):
            changeCursor = 'ibeam'

    return changeCursor

  def initiate(self, surface: pg.Surface):
    self.surface = surface

    self.locked = False



import pygame as pg
import pg_extended as pgx

class Utility:
  def changeTitle(self, title: str):
    self.title = title
    pg.display.set_caption(self.title)

  def screenResized(self) -> bool:
    if not self.running:
      return None

    tmpSW = self.screen.get_width()
    tmpSH = self.screen.get_height()

    if (self.screenWidth != tmpSW) or (self.screenHeight != tmpSH):
      self.screenWidth, self.screenHeight = self.screen.get_width(), self.screen.get_height()
      return True

    return False

  def setViewPort(self, viewPort: pgx.Game.ViewPort):
    self.viewPort = viewPort

    if self.running:
      self.viewPort.initiate(self.screen, self.activeScene)

  def resetUI(self):
    if not self.running:
      return None

    for dvKey in self.lazyDynamicValues:
      self.lazyDynamicValues[dvKey].resolveValue()

    for avKey in self.customAnimatedValues:
      self.customAnimatedValues[avKey].updateRestingPos()

    if self.customUpdateProcess is not None:
      self.customUpdateProcess()

    if self.activeScene is not None:
      self.activeScene.lazyUpdate()

    if self.viewPort is not None:
      self.viewPort.update()
      self.viewPort.renderScene()

    for systemID in self.systemZ:
      if systemID in self.activeSystems:
        self.activeSystems[systemID].lazyUpdate()



import types
import pygame as pg

type Background = pg.Color | pg.Surface

type TileIdentifier = tuple[int, int] | tuple[int, int, int] | tuple[int, int, int, int] | pg.Color | str

type CallableLike = types.FunctionType | types.BuiltinFunctionType | types.MethodType



from typing import Literal
from copy import copy
import pygame as pg
from pg_extended.Core import DynamicValue
from pg_extended.Core import CallbackSet
from pg_extended.Types import Background
from pg_extended.UI.Elements.Section import Section
from pg_extended.UI.Elements.TextBox import TextBox

class Button:
  def __init__(self, textBox: TextBox, callback: CallbackSet = None, border: int = 0, borderCol: Background = None, pressedBG: Background = None, pressedBorderBG: Background = None):
    self.textBox = textBox
    self.callback = callback
    self.border = border

    self.section = textBox.section

    self.pressedBG = pressedBG
    self.pressedBorderBG = pressedBorderBG

    self.defaultBG = copy(self.section.background)
    self.defaultBorderBG = borderCol

    self.borderSection = None

    if self.border > 0:
      self.borderSection = Section(
        {
          'x': DynamicValue(lambda: self.section.x - self.border),
          'y': DynamicValue(lambda: self.section.y - self.border),
          'width': DynamicValue(lambda: self.border * 2 + self.section.width),
          'height': DynamicValue(lambda: self.border * 2 + self.section.height)
        }, self.defaultBorderBG, self.section.borderRadius
      )

    self.active = True
    self.activeUpdate = True
    self.activeEvents = True
    self.activeDraw = True

    self.pressed = False

    self.lazyUpdate = True
    self.lazyUpdateOverride = False

  def switchBG(self):
    if self.pressed:
      if self.pressedBG:
        self.section.background = self.pressedBG

      if self.borderSection and self.pressedBorderBG:
        self.borderSection.background = self.pressedBorderBG

    else:
      self.section.background = self.defaultBG

      if self.borderSection:
        self.borderSection.background = self.defaultBorderBG

    self.section.update()

    if self.borderSection:
      self.borderSection.update()

  def handleCallback(self, event: pg.Event, eventType: Literal['mouseUp'] | Literal['mouseDown']):
    if self.callback is None:
      return None

    if eventType == 'mouseUp' and not self.section.rect.collidepoint(event.pos):
      return None

    self.callback.call(eventType)

  def checkEvent(self, event: pg.Event) -> bool | None:
    if not (self.active and self.activeEvents):
      return None

    if event.type == pg.MOUSEBUTTONDOWN and event.button == 1 and self.section.rect.collidepoint(event.pos):
      self.pressed = True

      self.switchBG()

      self.handleCallback(event, 'mouseDown')

      return True

    if event.type == pg.MOUSEBUTTONUP and self.pressed:
      self.pressed = False

      self.switchBG()

      self.handleCallback(event, 'mouseUp')

      return True

    return False

  def update(self):
    if not (self.active and self.activeUpdate):
      return None

    self.textBox.update()

    if self.border > 0:
      self.borderSection.update()

  def draw(self, surface: pg.Surface):
    if not (self.active and self.activeDraw):
      return None

    if self.border > 0:
      self.borderSection.draw(surface)

    self.textBox.draw(surface, True)



from json import load
import traceback
import pygame as pg
from pg_extended.Types import TileIdentifier
from pg_extended.Game.Elements import Entity

class Level:
  def __init__(self, numTilesX: int, numTilesY: int, tileWidth: float, tileHeight: float, tilesMatrixJsonURL: str):
    self.numTilesX = numTilesX
    self.numTilesY = numTilesY
    self.tileWidth = tileWidth
    self.tileHeight = tileHeight
    self.tilesMatrixJsonURL = tilesMatrixJsonURL

    self.tileLayers = []
    self.generateTilesMatrix()

    self.locked = True
    self.activeDraw = True
    self.surface: pg.Surface = None
    self.scene: 'Scene' = None # type: ignore
    self.entities: list[Entity] = []

    self.width = tileWidth * numTilesX
    self.height = tileHeight * numTilesY

  def generateTilesMatrix(self):
    try:
      with open(self.tilesMatrixJsonURL, 'r') as f:
        rawJson = load(f)
        layersRaw = rawJson['layers']

        for layer in layersRaw:
          self.tileLayers.append([])
          for row in layer:
            self.tileLayers[-1].append([])
            for tile in row:
              if tile is None:
                self.tileLayers[-1][-1].append(None)
              elif isinstance(tile[1], str):
                self.tileLayers[-1][-1].append((rawJson['atlases'][tile[0]], tile[1]))
              else:
                self.tileLayers[-1][-1].append((rawJson['atlases'][tile[0]], (*tile[1],)))

    except Exception as e:
      print(e)
      traceback.print_exc()

  def recalcDim(self):
    self.width = self.tileWidth * self.numTilesX
    self.height = self.tileHeight * self.numTilesY

  def renderLevelSurface(self):
    self.recalcDim()

    self.surface = pg.Surface((self.width, self.height), pg.SRCALPHA)

    for layer in self.tileLayers:
      y = -1
      for row in layer:
        y += 1
        x = -1
        for tile in row:
          x += 1

          if tile is None: continue

          atlasID, tileID = tile[0], tile[1]

          currentTile = pg.transform.scale(self.scene.textureAtlases[atlasID].getTile(tileID), (self.tileWidth, self.tileHeight))

          if currentTile is None: continue

          tilePos = (self.tileWidth * x, self.tileHeight * y)

          self.surface.blit(currentTile, tilePos)

  def updateTile(self, poses: list[tuple[int, int]] | tuple[tuple[int, int]], tiles: list[tuple[int, TileIdentifier]] | tuple[tuple[int, TileIdentifier]]):
    for i in range(len(poses)):
      x, y = poses[i]
      atlasID = tiles[i][0]
      tileID = tiles[i][1]

      tile = self.scene.textureAtlases[atlasID].getTile(tileID)

      if tile is None: continue

      tile = pg.transform.scale(tile, (self.tileWidth, self.tileHeight))

      tilePos = (self.tileWidth * x, self.tileHeight * y)

      self.surface.blit(tile, tilePos)

  def addEntity(self, entity: Entity):
    self.entities.append(entity)

  def initiate(self, scene: 'Scene') -> bool: # type: ignore
    try:
      self.scene = scene
      self.renderLevelSurface()

      for entity in self.entities:
        entity.initiate(self, scene)

      self.locked = False
      return True
    except Exception as e:
      print(e)
      traceback.print_exc()
      return False



from copy import copy
from pg_extended.UI.Elements import *

class CopyElement:
  @staticmethod
  def copyElement(element: UIElement) -> UIElement:
    # from pg_extended.UI import Section, Circle, TextBox, Button, Slider, TextInput

    if isinstance(element, Section):
      return CopyElement.copySection(element)
    elif isinstance(element, Circle):
      return CopyElement.copyCircle(element)
    elif isinstance(element, TextBox):
      return CopyElement.copyTextBox(element)
    elif isinstance(element, Button):
      return CopyElement.copyButton(element)
    elif isinstance(element, Toggle):
      return CopyElement.copyToggle(element)
    elif isinstance(element, Slider):
      return CopyElement.copySlider(element)
    elif isinstance(element, TextInput):
      return CopyElement.copyTextInput(element)
    else:
      raise ValueError('Unsupported element type for copying')

  @staticmethod
  def copySection(section: Section) -> Section:
    return Section(
      copy(section.dimensions),
      section.background,
      section.borderRadius,
      section.backgroundSizeType,
      section.backgroundPosition,
      section.backgroundSizePercent
    )

  @staticmethod
  def copyCircle(circle: Circle) -> Circle:
    return Circle(
      copy(circle.dimensions),
      circle.background,
      circle.backgroundSizeType,
      circle.backgroundSizePercent
    )

  @staticmethod
  def copyTextBox(textBox: TextBox) -> TextBox:
    element = TextBox(
      CopyElement.copySection(textBox.section),
      textBox.text,
      textBox.fontPath,
      textBox.textColor,
      textBox.fontSize
    )

    element.drawSectionDefault = textBox.drawSectionDefault
    element.alignTextHorizontal = textBox.alignTextHorizontal
    element.alignTextVertical = textBox.alignTextVertical
    element.paddingLeft = textBox.paddingLeft
    element.paddingRight = textBox.paddingRight
    element.active = textBox.active
    element.activeDraw = textBox.activeDraw
    element.activeUpdate = textBox.activeUpdate
    element.lazyUpdate = textBox.lazyUpdate
    element.lazyUpdateOverride = textBox.lazyUpdateOverride

    return element

  @staticmethod
  def copyButton(button: Button) -> Button:
    return Button(
      CopyElement.copySection(button.section),
      button.pressedBackground,
      button.borderColor,
      button.borderColorPressed,
      button.textBox.text,
      button.textBox.fontPath,
      button.textBox.textColor,
      button.border,
      button.callback
    )

  @staticmethod
  def copyToggle(toggle: Toggle) -> Toggle:
    return Toggle(
      CopyElement.copySection(toggle.section),
      toggle.indicatorColor,
      toggle.borderColor,
      toggle.borderColorToggled,
      toggle.border,
      toggle.callback
    )

  @staticmethod
  def copySlider(slider: Slider) -> Slider:
    return Slider(
      slider.orientation,
      CopyElement.copySection(slider.section),
      CopyElement.copySection(slider.dragElement) if slider.dragElementType == 'section' else CopyElement.copyCircle(slider.dragElement),
      slider.valueRange,  
      slider.scrollSpeed,
      slider.filledSliderBackground,
      slider.onChangeInfo,
      slider.hoverToScroll
    )

  @staticmethod
  def copyTextInput(textInput: TextInput) -> TextInput:
    return TextInput(
      CopyElement.copySection(textInput.section),
      textInput.fontPath,
      textInput.textColor,
      textInput.max,
      textInput.placeholder,
      textInput.placeholderTextColor,
      textInput.border,
      textInput.borderColor,
      textInput.focusBorderColor,
      textInput.focusBackground,
      textInput.onChangeInfo,
      textInput.alignTextHorizontal,
      textInput.alignTextVertical
    )



import time
from copy import copy
from pg_extended.Core.Base.DynamicValue import DynamicValue

INTERPOLATION_TYPES = ['linear', 'easeIn', 'easeOut', 'easeInOut', 'custom']
DEFAULT_POS_VALS = ['start', 'end']

type valuesType = list[DynamicValue | AnimatedValue | int | float] | list[DynamicValue | AnimatedValue | int | float]

class AnimatedValue:
  def __init__(self, values: valuesType, duration: float, defaultPos: str = 'start', interpolation: str = 'linear', callback: callable = None, customInterpolation: callable = None):
    if len(values) < 2:
      raise ValueError("Animator requires a minimum of two values to animate between.")

    if not interpolation in INTERPOLATION_TYPES:
      raise ValueError(f'Invalid interpolation type: {interpolation}. Must be one of: {INTERPOLATION_TYPES}')

    if interpolation == 'custom' and customInterpolation is None:
      raise ValueError('Custom interpolation function must be provided when using "custom" interpolation type.')

    if not defaultPos in DEFAULT_POS_VALS:
      raise ValueError(f'Invalid defaultPos: {defaultPos}. Must be one of: {DEFAULT_POS_VALS}')

    self.values = values
    self.rawValues: list[int | float] = []
    self.duration = duration
    self.interpolation = interpolation
    self.callback = callback
    self.defaultPos = defaultPos

    self.updateValues()

    if self.defaultPos == 'start':
      self.value = self.rawValues[0]
    else:
      self.value = self.rawValues[-1]

    self.animStart: float = None
    self.reverse: bool = False
    self.repeats: int = 0
    self.alternate: bool = False
    self.hasPlayedOnce: bool = False
    self.delay: int = 0

    if self.interpolation == 'linear':
      self.interpolationStep = self.linear
    elif self.interpolation == 'easeIn':
      self.interpolationStep = self.easeIn
    elif self.interpolation == 'easeOut':
      self.interpolationStep = self.easeOut
    elif self.interpolation == 'easeInOut':
      self.interpolationStep = self.easeInOut
    elif self.interpolation == 'custom':
      self.interpolationStep = customInterpolation

  @staticmethod
  def linear(start: float, end: float, t: float) -> float:
    if t <= 0:
      return start
    elif t >= 1:
      return end

    return start + (end - start) * t

  @staticmethod
  def easeIn(start: float, end: float, t: float) -> float:
    if t <= 0:
      return start
    elif t >= 1:
      return end

    t = t ** 2

    return start + (end - start) * t

  @staticmethod
  def easeOut(start: float, end: float, t: float) -> float:
    if t <= 0:
      return start
    elif t >= 1:
      return end

    t = 1 - (1 - t) ** 2

    return start + (end - start) * t

  @staticmethod
  def easeInOut(start: float, end: float, t: float) -> float:
    if t <= 0:
      return start
    elif t >= 1:
      return end

    t = t**3 * (t * (t * 6 - 15) + 10)

    return start + (end - start) * t

  # get raw values from animated / dynamic values
  def updateValues(self):
    self.rawValues = []

    for value in self.values:
      if isinstance(value, (DynamicValue, AnimatedValue)):
        value.resolveValue()
        self.rawValues.append(value.value)
      else:
        self.rawValues.append(value)

  # get an interpolated value from normalized t
  def interpolate(self, t: float):
    if t <= 0:
      return self.rawValues[0]
    elif t >= 1:
      return self.rawValues[-1]

    processingVals = copy(self.rawValues)

    while len(processingVals) > 1:
      tmp = []

      for i in range(len(processingVals) - 1):
        tmp.append(
          self.interpolationStep(processingVals[i], processingVals[i + 1], t)
        )

      processingVals = tmp

    self.value = processingVals[0]

  # calculate current animation time, get normalized t, call .interpolate() etc..
  # most importantly this is the function you need to call to update the animated value
  def resolveValue(self):
    if self.animStart is None:
      self.updateRestingPos()
      return

    elapsedTime = ((time.perf_counter() * 1000) - self.animStart) - self.delay

    self.updateValues()

    if elapsedTime >= self.duration:
      self.finishAnim()
    else:
      if self.reverse:
        t = 1 - (elapsedTime / self.duration)
      else:
        t = elapsedTime / self.duration

      self.interpolate(t)

  # handle animation ends, repeats, callbacks used by .resolveValue()
  def finishAnim(self):
    if self.reverse:
      self.value = self.rawValues[0]
    else:
      self.value = self.rawValues[-1]

    self.animStart = None

    self.hasPlayedOnce = True

    if self.repeats > 0:
      self.repeats -= 1
      self.trigger(self.reverse, self.repeats, self.alternate)
      return None

    if self.repeats == -1:
      self.trigger(self.reverse, self.repeats, self.alternate)
      return None

    if self.callback is not None:
      self.callback()

  # updates the value to a default idle position when no animation is playing
  def updateRestingPos(self):
    self.updateValues()

    A = self.hasPlayedOnce
    B = self.defaultPos == 'start'
    C = self.reverse

    '''
    this let's the system decide the resting position with only the default position when it hasn't ran yet
    once it has ran, the reverse value also has an effect on the resting value

    if self.hasPlayedOnce:
      if self.defaultPos == 'start':
        if self.reverse:
          self.value = self.values[0].value
        else:
          self.value = self.values[-1].value
      else:
        if self.reverse:
          self.value = self.values[-1].value
        else:
          self.value = self.values[0].value
    else:
      if self.defaultPos == 'start':
        self.value = self.values[0].value
      else:
        self.value = self.values[-1].value

    condition for choosing first value:
    (A and B and C) or (A and not B and not C) or (not A and B)
    condition for choosing last value:
    (A and B and not C) or (A and not B and C) or (not A and not B)

    compacted:
    (A and (B == C)) or (not A and B)

    maximum performance version:
    (B and not A) or (A and (B == C))
    '''

    pickStart = (B and not A) or (A and (B == C))

    self.value = self.rawValues[0] if pickStart else self.rawValues[-1]

  # triggers animation
  def trigger(self, reverse: bool = False, repeats: int = 0, alternate: bool = False, delay: int = 0):
    self.animStart = time.perf_counter() * 1000

    self.repeats = repeats
    self.alternate = alternate
    self.delay = delay

    if self.hasPlayedOnce:
      if self.alternate:
        self.reverse = not self.reverse
      else:
        self.reverse = reverse
    else:
      self.reverse = reverse

  # stops all animations, resets repeats and instantly snaps the value to a resting position
  def terminate(self):
    if self.animStart is None:
      return None

    self.animStart = None
    self.repeats = 0

    if self.reverse:
      self.value = self.rawValues[0]
    else:
      self.value = self.rawValues[-1]



import pygame as pg

class ImgManipulation:
  # deforms the image to perfectly fit in the container
  @staticmethod
  def squish(image: pg.Surface, containerSize: tuple[int | float, int | float] | list[int | float], smoothscale: bool = True, scalePercent: int | None = 100) -> pg.Surface:
    if smoothscale:
      return pg.transform.smoothscale(
        image,
        (
          containerSize[0] * (scalePercent / 100),
          containerSize[1] * (scalePercent / 100)
        )
      )

    return pg.transform.scale(
      image,
      (
        containerSize[0] * (scalePercent / 100),
        containerSize[1] * (scalePercent / 100)
      )
    )

  # resizes the image to the smallest possible fit while preserving the original aspect ratio
  @staticmethod
  def fit(image: pg.Surface, containerSize: tuple[int | float, int | float] | list[int | float], smoothscale: bool = True, scalePercent: int | None = 100) -> pg.Surface:
    containerWidth, containerHeight = containerSize

    imageWidth, imageHeight = image.get_width(), image.get_height()

    scale = min(containerWidth / imageWidth, containerHeight / imageHeight) * (scalePercent / 100)

    newWidth = int(imageWidth * scale)
    newHeight = int(imageHeight * scale)

    if smoothscale:
      return pg.transform.smoothscale(image, (newWidth, newHeight))

    return pg.transform.scale(image, (newWidth, newHeight))

  # resizes the image to the largest possible fit while preserving the original aspect ratio
  @staticmethod
  def fill(image: pg.Surface, containerSize: tuple[int | float, int | float] | list[int | float], smoothscale: bool = True, scalePercent: int | None = 100) -> pg.Surface:
    containerWidth, containerHeight = containerSize

    imageWidth, imageHeight = image.get_width(), image.get_height()

    scale = max(containerWidth / imageWidth, containerHeight / imageHeight) * (scalePercent / 100)

    newWidth = int(imageWidth * scale)
    newHeight = int(imageHeight * scale)

    if smoothscale:
      return pg.transform.smoothscale(image, (newWidth, newHeight))

    return pg.transform.scale(image, (newWidth, newHeight))

  # returns an image with it's borders rounded.
  @staticmethod
  def roundImage(img: pg.Surface, radius: float) -> pg.Surface:
    w, h = img.get_size()

    mask = pg.Surface((w, h), pg.SRCALPHA)

    pg.draw.rect(mask, (255, 255, 255, 255), (0, 0, w, h), border_radius=radius)

    result = pg.Surface((w, h), pg.SRCALPHA)

    result.blit(img, (0, 0))
    result.blit(mask, (0, 0), special_flags=pg.BLEND_RGBA_MULT)

    return result

  # make a 2xn or nx2 pixel surface with provided colors that can later be smoothscaled to get a color gradient
  @staticmethod
  def getGradient(colors: list[pg.Color | tuple[int, int, int] | tuple[int, int, int, int]], sizes: list[int], direction: str, thickness: int = 2) -> pg.Surface:
    if direction in ('up', 'left'):
      colors.reverse()
      sizes.reverse()

    if direction in ('up', 'down'):
      w = thickness
      h = sum(sizes)
    else:
      w = sum(sizes)
      h = thickness

    surface = pg.Surface((w, h), pg.SRCALPHA)

    if direction in ('up', 'down'):
      for i in range(len(colors)):
        pg.draw.rect(surface, colors[i], (0, sum(sizes[:i]), thickness, sizes[i]), 0)
    else:
      for i in range(len(colors)):
        pg.draw.rect(surface, colors[i], (sum(sizes[:i]), 0, sizes[i], thickness), 0)

    return surface



import pygame as pg
from pg_extended.Game.Elements import *

class Scene:
  def __init__(self):
    self.locked = True
    self.surface: pg.Surface = None

    self.elements: dict[str, GameElement] = {}
    self.textureAtlases: dict[str, TextureAtlas] = {}
    self.spriteAnimations: dict[str, SpriteAnimation] = {}
    self.levels: dict[str, Level] = {}
    self.activeLevel: Level = None
    self.players: dict[str, Player] = {}

  def addElement(self, element: GameElement, elementID: str):
    if elementID in self.elements:
      raise ValueError(f'An element with ID: {elementID} already exists, please enter a unique ID.')

    self.elements[elementID] = element

    if isinstance(element, TextureAtlas):
      self.textureAtlases[elementID] = element
    elif isinstance(element, SpriteAnimation):
      self.spriteAnimations[elementID] = element
    elif isinstance(element, Level):
      self.levels[elementID] = element
    elif isinstance(element, Player):
      self.players[elementID] = element

  def lazyUpdate(self):
    pass

  def update(self):
    if self.locked:
      print('Scene is currently locked')
      return None

    for level in self.levels.values():
      for entity in level.entities:
        if entity.animating:
          entity.update()

  def handleEvents(self, event: pg.Event):
    pass

  def activateLevel(self, levelID: str):
    if not levelID in self.levels:
      print(f'Level with ID {levelID} does not exist, plase enter an existing level ID')
      return None

    self.activeLevel = self.levels[levelID]

  def deactivateLevel(self):
    self.activeLevel = None

  def initiate(self, surface: pg.Surface):
    self.surface = surface

    for atlas in self.textureAtlases.values():
      atlas.generateTiles()

    for spriteAnimation in self.spriteAnimations.values():
      spriteAnimation.initiate(self)

    levelInitializationSuccess = [level.initiate(self) for level in self.levels.values()]

    if all(levelInitializationSuccess):
      self.locked = False



from pg_extended.Game.Elements.TextureAtlas import TextureAtlas
from pg_extended.Game.Elements.SpriteAnimation import SpriteAnimation
from pg_extended.Game.Elements.Entity import Entity
from pg_extended.Game.Elements.Player import Player
from pg_extended.Game.Elements.Level import Level

from typing import Union

type GameElement = Union[TextureAtlas, SpriteAnimation, Entity, Player, Level]



from math import floor
import pygame as pg
from pg_extended.Types import TileIdentifier
from pg_extended.Core import DynamicValue, AnimatedValue
from pg_extended.Game.Elements.SpriteAnimation import SpriteAnimation
from pg_extended.Game.Elements import Level

class Entity:
  def __init__(self, name: str, x: float, y: float, width: float, height: float, defaultSpriteDetails: tuple[str, TileIdentifier]):
    self.name = name
    self.x = x
    self.y = y
    self.width = width
    self.height = height
    self.defaultSpriteAtlasID = defaultSpriteDetails[0]
    self.defaultSpriteID = defaultSpriteDetails[1]

    self.locked: bool = True
    self.animating: bool = False
    self.xPX: float = 0.0
    self.yPX: float = 0.0
    self.widthPX: float = 0.0
    self.heightPX: float = 0.0
    self.animationFrames: int = 1
    self.animations: list[dict[str, tuple[float, SpriteAnimation]]] = []
    self.currentAnimation: SpriteAnimation = None
    self.defaultSprite: pg.Surface = None
    self.sprite: pg.Surface = None

    self.animationInterpolator: AnimatedValue = AnimatedValue(
      [DynamicValue(0), DynamicValue(self, 'animationFrames')],
      1,
      'start',
      'linear',
      self.terminateAnimation
    )

  def initiate(self, level: Level, scene: 'Scene'): # type: ignore
    self.level = level
    self.xPX = self.x * level.tileWidth
    self.yPX = self.y * level.tileHeight
    self.widthPX = self.width * level.tileWidth
    self.heightPX = self.height * level.tileHeight
    self.defaultSprite = scene.textureAtlases[self.defaultSpriteAtlasID].getTile(self.defaultSpriteID)
    self.sprite = self.defaultSprite
    self.locked = False

  def update(self):
    if self.locked: return

    if self.animating:
      self.animationInterpolator.resolveValue()

      spriteIndex = floor(self.animationInterpolator.value)

      if spriteIndex >= self.animationFrames:
        spriteIndex = self.animationFrames - 1
      elif spriteIndex < 0:
        spriteIndex = 0

      self.sprite = self.currentAnimation.sprites[spriteIndex]
    else:
      self.sprite = self.defaultSprite

  def setAnimations(self, animations: dict[str, SpriteAnimation]):
    self.animations = animations

  def triggerAnimation(self, animationName: str, duration: float, reverse: bool = False, repeats: int = 0, alternate: bool = False):
    self.currentAnimation = self.animations[animationName]
    self.animationFrames = len(self.currentAnimation.sprites)
    self.animating = True
    self.animationInterpolator.duration = duration
    self.animationInterpolator.trigger(reverse, repeats, alternate)

  def terminateAnimation(self):
    self.animating = False
    self.animationInterpolator.terminate()



from typing import Literal
import pygame as pg
from pg_extended.Types import Background
from pg_extended.Util import ImgManipulation
from pg_extended.Core import DynamicValue, AnimatedValue, RectArea

VALID_SIZE_TYPES = ('fit', 'fill', 'squish', 'none')

type NumValue = DynamicValue | AnimatedValue | int | float

class Section(RectArea):
  def __init__(self, dimensions: dict[str, NumValue], background: Background, borderRadius: float | None = 0, backgroundSizeType: str | None = 'fit', backgroundPosition: str | None = 'center', backgroundSizePercent: int | None = 100):
    self.background = background
    self.midProcessBG = None
    self.drawReady = None

    self.backgroundSizeType = backgroundSizeType
    self.backgroundSizePercent = backgroundSizePercent
    self.backgroundPosition = backgroundPosition
    self.backgroundOffset = [0, 0]

    self.borderRadius = borderRadius

    self.backgroundSmoothScale = True

    self.active = True
    self.activeDraw = True
    self.activeUpdate = True
    self.lazyUpdate = True
    self.lazyUpdateOverride = False

    if not self.backgroundSizeType in VALID_SIZE_TYPES:
      raise ValueError(f'Invalid \"backgroundSizeType\" value, must be one of the following values: {VALID_SIZE_TYPES}')

    super().__init__(dimensions)

    self.update()

  def resizeBackground(self, getFrom: Literal['raw', 'mid', 'final'], setTo: Literal['raw', 'mid', 'final']):
    if not isinstance(self.background, pg.Surface):
      return None

    if getFrom == 'raw':
      sourceImage = self.background
    elif getFrom == 'mid':
      sourceImage = self.midProcessBG
    else:
      sourceImage = self.drawReady

    if self.backgroundSizeType == 'fit':
      result = ImgManipulation.fit(sourceImage, (self.width, self.height), self.backgroundSmoothScale, self.backgroundSizePercent)
    elif self.backgroundSizeType == 'fill':
      result = ImgManipulation.fill(sourceImage, (self.width, self.height), self.backgroundSmoothScale, self.backgroundSizePercent)
    elif self.backgroundSizeType == 'squish':
      result = ImgManipulation.squish(sourceImage, (self.width, self.height), self.backgroundSmoothScale, self.backgroundSizePercent)
    elif not self.backgroundSizePercent == 100:
      result = ImgManipulation.fit(sourceImage, (self.background.get_width(), self.background.get_height()), self.backgroundSmoothScale, self.backgroundSizePercent)
    else:
      result = sourceImage

    if setTo == 'raw':
      self.background = result
    elif setTo == 'mid':
      self.midProcessBG = result
    else:
      self.drawReady = result

  def applyRadiusToBackground(self, getFrom: Literal['raw', 'mid', 'final'], setTo: Literal['raw', 'mid', 'final']):
    if not isinstance(self.background, pg.Surface):
      return None

    if getFrom == 'raw':
      sourceImage = self.background
    elif getFrom == 'mid':
      sourceImage = self.midProcessBG
    else:
      sourceImage = self.drawReady

    if self.borderRadius is None or self.borderRadius <= 0:
      result = sourceImage
    else:
      result = ImgManipulation.roundImage(sourceImage, self.borderRadius)

    if setTo == 'raw':
      self.background = result
    elif setTo == 'mid':
      self.midProcessBG = result
    else:
      self.drawReady = result

  def setBackgroundPos(self, source: Literal['raw', 'mid', 'final']):
    if not isinstance(self.background, pg.Surface):
      return None

    if source == 'raw':
      sourceImage = self.background
    elif source == 'mid':
      sourceImage = self.midProcessBG
    else:
      sourceImage = self.drawReady

    # set x position
    if self.backgroundPosition.endswith('left'):
      self.imageX = self.x
    elif self.backgroundPosition.endswith('center'):
      self.imageX = self.x + ((self.width - sourceImage.get_width()) / 2)
    elif self.backgroundPosition.endswith('right'):
      self.imageX = self.x + (self.width - sourceImage.get_width())

    # set y position
    if self.backgroundPosition.startswith('top'):
      self.imageY = self.y
    elif self.backgroundPosition.startswith('center'):
      self.imageY = self.y + ((self.height - sourceImage.get_height()) / 2)
    elif self.backgroundPosition.startswith('bottom'):
      self.imageY = self.y + (self.height - sourceImage.get_height())

    # apply offset
    self.imageX += self.backgroundOffset[0]
    self.imageY += self.backgroundOffset[1]

  def createTransparentSurface(self, setTo: Literal['raw', 'mid', 'final']):
    if not isinstance(self.background, pg.Color) or self.background.a == 255:
      return None

    if setTo == 'raw':
      self.background = pg.Surface(self.rect.size, pg.SRCALPHA)
      pg.draw.rect(self.background, self.background, (0, 0, self.width, self.height), border_radius=self.borderRadius)
    elif setTo == 'mid':
      self.midProcessBG = pg.Surface(self.rect.size, pg.SRCALPHA)
      pg.draw.rect(self.midProcessBG, self.background, (0, 0, self.width, self.height), border_radius=self.borderRadius)
    else:
      self.drawReady = pg.Surface(self.rect.size, pg.SRCALPHA)
      pg.draw.rect(self.drawReady, self.background, (0, 0, self.width, self.height), border_radius=self.borderRadius)

  def update(self):
    if not (self.active and self.activeUpdate):
      return None

    super().update()

    if isinstance(self.background, pg.Surface):
      self.resizeBackground('raw', 'mid')
      self.applyRadiusToBackground('mid', 'final')
      self.setBackgroundPos('final')

    elif isinstance(self.background, pg.Color) and self.background.a < 255:
      self.createTransparentSurface('final')

    elif isinstance(self.background, pg.Color):
      self.drawReady = self.background

  def draw(self, surface: pg.Surface):
    if not (self.active and self.activeDraw):
      return None

    if isinstance(self.background, pg.Surface):
      surface.blit(self.drawReady, (self.imageX, self.imageY))
    elif isinstance(self.background, pg.Color):
      if self.background.a < 255:
        surface.blit(self.drawReady, self.rect.topleft)
      else:
        pg.draw.rect(surface, self.background, self.rect, border_radius=self.borderRadius)



from pg_extended.Game import Elements
from pg_extended.Game.Scene import Scene
from pg_extended.Game.ViewPort import ViewPort

from pg_extended.Game.Elements import *

__all__ = [
  'TextureAtlas',
  'SpriteAnimation',
  'Entity',
  'Player',
  'Level',
  'Scene',
  'ViewPort'
]



from pg_extended.UI import Elements
from pg_extended.UI import Compounds

from pg_extended.UI.Elements import *
from pg_extended.UI.Compounds import *

from pg_extended.UI.System import System
from pg_extended.UI.CopyElement import CopyElement

__all__ = [
  'Section',
  'Circle',
  'TextBox',
  'TextInput',
  'Button',
  'Toggle',
  'Slider',
  'System',
  'List',
  'UIElement',
  'UICompound',
  'CopyElement'
]



import pygame as pg
from pg_extended.Core import DynamicValue, AnimatedValue, Callback
from pg_extended.UI.Elements.Section import Section

class Toggle:
  def __init__(self, section: Section, indicatorColor: pg.Color, borderColor: pg.Color, borderColorToggled: pg.Color, border: int = 0, callback: Callback = None):
    self.section = section
    self.defaultBackground = section.background
    self.indicatorColor = indicatorColor
    self.borderColor = borderColor
    self.borderColorToggled = borderColorToggled
    self.border = border
    self.callback = callback

    self.borderRect = pg.Rect(self.section.x - border, self.section.y - border, self.section.width + (border * 2), self.section.height + (border * 2))

    self.toggled = False
    self.active = True
    self.activeDraw = True
    self.activeUpdate = True
    self.activeEvents = True
    self.lazyUpdate = True
    self.lazyUpdateOverride = False

    self.innerBoxPadding = .1
    self.innerBox = pg.Rect(0, 0, 0, 0)

    self.innerBoxAnim = AnimatedValue(
      [
        DynamicValue(lambda v: v.section.x + (v.section.width * v.innerBoxPadding), args={'v': self}),
        DynamicValue(lambda v: v.section.x + (v.section.width / 2), args={'v': self})
      ], 70, 'start', 'linear', self.animationCallback
    )

    self.update()

  def animationCallback(self):
    self.lazyUpdateOverride = False
    self.update()

  def updateInnerBox(self):
    self.innerBoxAnim.resolveValue()

    newX = self.innerBoxAnim.value
    newY = self.section.y  + (self.section.height * self.innerBoxPadding)
    newW = (self.section.width / 2) * (1 - (self.innerBoxPadding * 2))
    newH = self.section.height * (1 - (self.innerBoxPadding * 2))

    self.innerBox.update(newX, newY, newW, newH)

  def checkEvent(self, event: pg.Event) -> bool | None:
    if not (self.active and self.activeEvents) or (self.innerBoxAnim.animStart is not None):
      return None

    if event.type == pg.MOUSEBUTTONDOWN and event.button == 1 and self.section.rect.collidepoint(event.pos):
      self.innerBoxAnim.trigger(self.toggled)

      self.lazyUpdateOverride = True

      self.toggled = not self.toggled

      if self.toggled:
        self.section.background = self.indicatorColor
      else:
        self.section.background = self.defaultBackground

      self.updateInnerBox()

      if self.callback is not None:
        self.callback.call({'value': self.toggled})

      return True

    return False

  def update(self):
    if not (self.active and self.activeUpdate):
      return None

    self.section.update()

    if self.toggled:
      self.section.background = self.indicatorColor
    else:
      self.section.background = self.defaultBackground

    newBorderX, newBorderY = self.section.x - self.border, self.section.y - self.border
    newBorderWidth, newBorderHeight = self.section.width + (self.border * 2), self.section.height + (self.border * 2)

    self.borderRect.update(newBorderX, newBorderY, newBorderWidth, newBorderHeight)

    self.updateInnerBox()

  def draw(self, surface: pg.Surface):
    if not (self.active and self.activeDraw):
      return None

    if self.border > 0:
      if self.toggled:
        pg.draw.rect(surface, self.borderColorToggled, self.borderRect, border_radius = self.section.borderRadius)
      else:
        pg.draw.rect(surface, self.borderColor, self.borderRect, border_radius = self.section.borderRadius)

    self.section.draw(surface)

    if self.toggled:
      pg.draw.rect(surface, self.defaultBackground, self.innerBox, border_radius = self.section.borderRadius)
    else:
      pg.draw.rect(surface, self.indicatorColor, self.innerBox, border_radius = self.section.borderRadius)



import time
import pyperclip
import pygame as pg
from pg_extended.Types import Background
from pg_extended.Core import DynamicValue, AnimatedValue, Callback
from pg_extended.UI.Elements.Section import Section
from pg_extended.UI.Elements.TextBox import TextBox

LINE_SPLIT_UNICODES = ' \t\u00A0\u2000\u200A\u3000'+',.;:!?\'\"(){}[]/\\|-_\n\r\f\v'

class TextInput:
  def __init__(self, section: Section, fontPath: str, textColor: pg.Color, max: int = -1, placeholder: str = None, placeholderTextColor: pg.Color = None, border: int = 0, borderColor: pg.Color = None, focusBorderColor: pg.Color = None, focusBackground: Background = None, callback: Callback = None, alignTextHorizontal: str = 'center', alignTextVertical: str = 'center'):
    self.section = section
    self.fontPath = fontPath
    self.textColor = textColor
    self.max = max
    self.placeholder = placeholder
    self.placeholderTextColor = placeholderTextColor
    self.border = border
    self.borderColor = borderColor
    self.focusBorderColor = focusBorderColor
    self.background = self.section.background
    self.focusBackground = focusBackground
    self.callback = callback
    self.alignTextHorizontal = alignTextHorizontal
    self.alignTextVertical = alignTextVertical

    self.inFocus = False
    self.typing = False
    self.active = True
    self.activeDraw = True
    self.activeUpdate = True
    self.activeEvents = True
    self.lazyUpdate = True
    self.lazyUpdateOverride = False
    self.inputText = ''
    self.lastEvent = ''
    self.lastKey = ''
    self.events = {}
    self.valueOnLastCallback = ''
    self.typingStart = 0
    self.lastAutoInputTime = 0
    self.autoInputDelay = 0.5
    self.autoInputInterval = 0.06
    self.dynamicAutoInputInterval = self.autoInputInterval
    self.autoInputSpeedIncrease = 0.8
    self.autoInputMinInterval = 0.01
    self.cursor: pg.Surface = None
    self.cursorX: float = 0

    self.cursorAlpha: AnimatedValue = AnimatedValue(
      [DynamicValue(255), DynamicValue(0)],
      500, 'start', 'easeIn'
    )

    if self.placeholderTextColor is None:
      self.placeholderTextColor = self.textColor

    if self.border > 0:
      self.borderRect = pg.Rect(self.section.x - border, self.section.y - border, self.section.width + (border * 2), self.section.height + (border * 2))

    self.textBox = TextBox(self.section, self.placeholder, self.fontPath, self.placeholderTextColor, False)

    self.textBox.alignTextHorizontal = alignTextHorizontal
    self.textBox.alignTextVertical = alignTextVertical

    self.textBox.paddingLeft = 2
    self.textBox.paddingRight = 2

    self._setupEvents()

    self.update()

  @staticmethod
  def getSplitText(text):
    splitArr = ['']

    for char in text:
      if char.isspace():
        if splitArr[-1].isspace():
          splitArr[-1] += char
        else:
          splitArr.append(char)
      elif char in LINE_SPLIT_UNICODES:
        splitArr.append(char)
      else:
        if splitArr[-1] and not splitArr[-1].isspace() and splitArr[-1] not in LINE_SPLIT_UNICODES:
          splitArr[-1] += char
        else:
          splitArr.append(char)

    if splitArr[0] == '':
      splitArr = splitArr[1:]

    return splitArr

  def _setupEvents(self):
    def unicode():
      self.inputText += self.lastKey

    def backspace():
      self.inputText = self.inputText[:-1]

    def ctrlBackspace():
      splitArr = self.getSplitText(self.inputText)

      self.inputText = ''.join(splitArr[:-1])

    def copy():
      pyperclip.copy(self.inputText)

    def paste():
      self.inputText += pyperclip.paste()

    self.events = {
      'unicode': unicode,
      'backspace': backspace,
      'ctrlBackspace': ctrlBackspace,
      'copy': copy,
      'paste': paste,
      'pass': lambda: None
    }

  def setTextBoxValue(self):
    if self.inputText == '':
      self.textBox.textColor = self.placeholderTextColor
      self.textBox.text = self.placeholder
    else:
      self.textBox.textColor = self.textColor
      self.textBox.text = self.inputText

  def handleCallback(self):
    if not self.active:
      return None

    if (not self.callback is None) and (not self.valueOnLastCallback == self.inputText):
      self.valueOnLastCallback = self.inputText

      self.callback.call({'value': self.inputText})

  def checkEvent(self, event: pg.Event) -> bool | None:
    if not (self.active and self.activeEvents):
      return None

    if event.type == pg.MOUSEBUTTONDOWN:
      if event.button == 1 and self.section.rect.collidepoint(event.pos):
        if not self.inFocus:
          self.inFocus = True
          self.lazyUpdateOverride = True

          self.cursorAlpha.trigger(False, -1, True)

          if self.focusBackground:
            self.section.background = self.focusBackground
            self.section.update()
      else:
        self.inFocus = False
        self.lazyUpdateOverride = False

        self.cursorAlpha.terminate()

        self.section.background = self.background
        self.section.update()

    elif self.inFocus and event.type == pg.KEYDOWN:
      self.typing = True
      self.typingStart = time.perf_counter()

      eventTriggered = False

      if event.mod & pg.KMOD_CTRL:
        if event.key == pg.K_BACKSPACE:
          eventTriggered = True
          self.lastEvent = 'ctrlBackspace'
        elif event.key == pg.K_c:
          eventTriggered = True
          self.lastEvent = 'copy'
        elif event.key == pg.K_v:
          eventTriggered = True
          self.lastEvent = 'paste'
        else:
          self.lastEvent = 'pass'
      else:
        eventTriggered = True
        if event.key == pg.K_BACKSPACE:
          self.lastEvent = 'backspace'
        else:
          self.lastKey = event.unicode
          self.lastEvent = 'unicode'

      if eventTriggered:
        self.events[self.lastEvent]()
        self.setTextBoxValue()
        self.textBox.update()

    elif event.type == pg.KEYUP:
      if self.typing:
        self.typing = False
        self.dynamicAutoInputInterval = self.autoInputInterval

        self.handleCallback()

  def update(self):
    if not (self.active and self.activeUpdate):
      return None

    newX, newY = self.section.x - self.border, self.section.y - self.border
    newWidth, newHeight = self.section.width + (self.border * 2), self.section.height + (self.border * 2)

    if self.inFocus:
      self.cursorAlpha.resolveValue()

      self.cursor = pg.Surface((2, self.textBox.textRect.height), pg.SRCALPHA)

      self.cursor.fill((self.textColor.r, self.textColor.g, self.textColor.b, self.cursorAlpha.value))

      spaceWidth, _ = self.textBox.font.size(' ')
      leftPaddingWidth = spaceWidth * self.textBox.paddingLeft
      rightPaddingWidth = spaceWidth * self.textBox.paddingRight
      textWidth, _ = self.textBox.font.size(self.inputText)

      if self.textBox.alignTextHorizontal == 'left':
        self.cursorX = self.section.x + leftPaddingWidth + textWidth
      elif self.textBox.alignTextHorizontal == 'center':
        self.cursorX = self.section.x + (self.section.width / 2) + ((textWidth + (leftPaddingWidth - rightPaddingWidth)) / 2)
      elif self.textBox.alignTextHorizontal == 'right':
        self.cursorX = self.section.x + (self.section.width - rightPaddingWidth)

    # auto rapid input on key hold
    if self.typing and (time.perf_counter() - self.typingStart > self.autoInputDelay):
      if time.perf_counter() - self.lastAutoInputTime > self.dynamicAutoInputInterval:
        if self.dynamicAutoInputInterval > self.autoInputMinInterval:
          self.dynamicAutoInputInterval *= self.autoInputSpeedIncrease

        self.lastAutoInputTime = time.perf_counter()

        self.events[self.lastEvent]()

        self.setTextBoxValue()

    self.textBox.update()

    if self.border > 0:
      self.borderRect.update(newX, newY, newWidth, newHeight)

  def draw(self, surface: pg.Surface):
    if not (self.active and self.activeDraw):
      return None

    if self.border > 0:
      if self.inFocus:
        pg.draw.rect(surface, self.focusBorderColor, self.borderRect, border_radius = self.section.borderRadius)
      else:
        pg.draw.rect(surface, self.borderColor, self.borderRect, border_radius = self.section.borderRadius)

    self.section.draw(surface)

    self.textBox.draw(surface)

    if self.inFocus:
      surface.blit(self.cursor, (self.cursorX, self.textBox.textRect.y))



from math import sqrt
import pygame as pg
from pg_extended.Types import Background
from pg_extended.Util import ImgManipulation
from pg_extended.Core import DynamicValue, CircleArea

VALID_SIZE_TYPES = ('fit', 'fill', 'squish', 'none')

type NumValue = DynamicValue | int | float

class Circle(CircleArea):
  def __init__(self, dimensions: dict[str, NumValue], background: Background, backgroundSizeType: str | None = 'fit', backgroundSizePercent: int | None = 100):
    self.background = background
    self.backgroundSizeType = backgroundSizeType
    self.backgroundSizePercent = backgroundSizePercent

    self.sqrt2 = sqrt(2)
    self.drawImage = None
    self.backgroundWidth = 0
    self.backgroundHeight = 0
    self.active = True
    self.activeDraw = True
    self.activeUpdate = True
    self.lazyUpdate = True
    self.lazyUpdateOverride = False

    if not self.backgroundSizeType in VALID_SIZE_TYPES:
      raise ValueError(f'Invalid \"backgroundSizeType\" value, must be one of the following values: {VALID_SIZE_TYPES}')

    super().__init__(dimensions)

    self.update()

  def update(self):
    if not (self.active and self.activeUpdate):
      return None

    super().update()

    if isinstance(self.background, pg.Surface):
      if self.backgroundSizeType == 'fit':
        self.drawImage = ImgManipulation.fit(self.background, (self.radius * self.sqrt2, self.radius * self.sqrt2), self.backgroundSizePercent)
      elif self.backgroundSizeType == 'fill':
        self.drawImage = ImgManipulation.fill(self.background, (self.radius * 2, self.radius * 2), self.backgroundSizePercent)
      else:
        self.drawImage = ImgManipulation.squish(self.background, (self.radius * 2, self.radius * 2), self.backgroundSizePercent)

      self.backgroundWidth = self.drawImage.get_width()
      self.backgroundHeight = self.drawImage.get_height()
    elif isinstance(self.background, pg.Color) and self.background.a < 255:
      self.drawImage = pg.Surface((self.radius * 2, self.radius * 2), pg.SRCALPHA)

      pg.draw.aacircle(self.drawImage, self.background, (self.radius, self.radius), self.radius)

  def draw(self, surface: pg.Surface):
    if not (self.active and self.activeDraw):
      return None

    if isinstance(self.background, pg.Surface):
      surface.blit(self.drawImage, (self.x - (self.backgroundWidth / 2), self.y - (self.backgroundHeight / 2)))
    elif isinstance(self.background, pg.Color):
      if self.background.a < 255:
        surface.blit(self.drawImage, (self.x - self.radius, self.y - self.radius))
      else:
        pg.draw.aacircle(surface, self.background, (self.x, self.y), self.radius)



import pygame as pg

class MainLoop:
  def updateLoop(self):
    self.handleEvents()

    if self.secondResize or self.screenResized():
      self.secondResize = not self.secondResize
      self.resetUI()

    self.currentFPS = self.clock.get_fps()

    self.screen.fill((0, 0, 0))

    for dvKey in self.customDynamicValues:
      self.customDynamicValues[dvKey].resolveValue()

    for avKey in self.customAnimatedValues:
      self.customAnimatedValues[avKey].resolveValue()

    if self.activeScene is not None:
      self.activeScene.update()

    for systemID in self.systemZ:
      if systemID in self.activeSystems:
        self.activeSystems[systemID].update()

    if self.customLoopProcess is not None:
      self.customLoopProcess()

    if self.viewPort is not None:
      self.viewPort.update()

      if not self.viewPort.lazyRender:
        self.viewPort.renderScene()

      self.viewPort.draw()

    for systemID in self.systemZ:
      if systemID in self.activeSystems:
        self.activeSystems[systemID].draw()

    if self.customDrawProcess is not None:
      self.customDrawProcess()

    self.firstUpdate = False
    pg.display.flip()
    self.clock.tick(self.fps)



from json import load
import pygame as pg
from pg_extended.Types import TileIdentifier

class TextureAtlas:
  def __init__(self, tilesetURL: str, tileWidth: int, tileHeight: int, paddingX: int = 0, paddingY: int = 0, tilestOffsetX: int = 0, tilestOffsetY: int = 0, namesJsonURL: str = None, sequences: dict[str, list[TileIdentifier] | tuple[TileIdentifier]] = None):
    self.tileWidth = tileWidth
    self.tileHeight = tileHeight
    self.paddingX = paddingX
    self.paddingY = paddingY
    self.tilestOffsetX = tilestOffsetX
    self.tilestOffsetY = tilestOffsetY
    self.namesJsonURL = namesJsonURL

    if namesJsonURL is not None:
      try:
        with open(namesJsonURL, 'r') as f:
          self.names = load(f)['names']
      except:
        self.names = None
    else:
      self.names = None

    self.sequences = sequences

    self.tileset = pg.image.load(tilesetURL)

    self.tilesX = (self.tileset.get_width() - tilestOffsetX) // (tileWidth + paddingX)
    self.tilesY = (self.tileset.get_height() - tilestOffsetY) // (tileHeight + paddingY)

    self.tiles = []
    self.namedTiles = {}
    self.sequencedTiles = {}

  def generateTiles(self):
    self.tileset.convert_alpha()

    for x in range(self.tilesX):
      self.tiles.append([])
      for y in range(self.tilesY):
        tileX = self.tilestOffsetX + (x * (self.tileWidth + self.paddingX))
        tileY = self.tilestOffsetY + (y * (self.tileHeight + self.paddingY))

        tileRect = pg.Rect(tileX, tileY, self.tileWidth, self.tileHeight)

        currentTile = pg.Surface(tileRect.size, pg.SRCALPHA)

        currentTile.blit(self.tileset, (0, 0), tileRect)

        self.tiles[x].append(currentTile)

    self.setNamedTiles()
    self.setSequencedTiles()

  def setNamedTiles(self):
    if self.names is None: return None

    y = -1
    for row in self.names:
      y += 1
      x = -1
      for name in row:
        x += 1
        self.namedTiles[name] = self.tiles[x][y]

  def setSequencedTiles(self):
    if self.sequences is None: return None

    for sequence in self.sequences:
      self.sequencedTiles[sequence] = []
      for tileIdentifier in self.sequences[sequence]:
        if isinstance(tileIdentifier, str):
          self.sequencedTiles[sequence].append(self.namedTiles[tileIdentifier])
        elif isinstance(tileIdentifier, tuple):
          if len(tileIdentifier) == 2:
            self.sequencedTiles[sequence].append(self.tiles[tileIdentifier[0]][tileIdentifier[1]])
          elif len(tileIdentifier) == 3 or len(tileIdentifier) == 4:
            surface = pg.Surface((self.tileWidth, self.tileHeight), pg.SRCALPHA)
            surface.fill(tileIdentifier)
            self.sequencedTiles[sequence].append(surface)
        elif isinstance(tileIdentifier, pg.Color):
          surface = pg.Surface((self.tileWidth, self.tileHeight), pg.SRCALPHA)
          surface.fill(tileIdentifier)
          self.sequencedTiles[sequence].append(surface)

  def getTile(self, identifier: TileIdentifier) -> pg.Surface | None:
    if isinstance(identifier, str):
      return self.namedTiles.get(identifier)
    elif isinstance(identifier, tuple):
      if len(identifier) == 2:
        return self.tiles[identifier[0]][identifier[1]]
      elif len(identifier) == 3 or len(identifier) == 4:
        surface = pg.Surface((self.tileWidth, self.tileHeight), pg.SRCALPHA)
        surface.fill(identifier)
        return surface
    elif isinstance(identifier, pg.Color):
      surface = pg.Surface((self.tileWidth, self.tileHeight), pg.SRCALPHA)
      surface.fill(identifier)
      return surface

    return None



from pg_extended.UI.Elements import UIElement
from pg_extended.UI.Elements import Circle
from pg_extended.UI.CopyElement import CopyElement
from pg_extended.Core import DynamicValue

class List:
  def __init__(self, listPos: dict[str, DynamicValue], listElement: UIElement, length: int, spacing: DynamicValue = None):
    self.listPos = listPos
    self.listElement = listElement
    self.length = length

    self.elements = []

    if spacing is None:
      self.spacing = DynamicValue(0)
    else:
      self.spacing = spacing

    for i in range(length):
      newElement = CopyElement.copyElement(listElement)

      if hasattr(newElement, 'section'):
        newElement.section.dimensions['x'] = self.listPos['x']
        newElement.section.dimensions['y'] = DynamicValue(self.getElementY, args={'index': i})
      elif hasattr(newElement, 'dimensions'):
        newElement.dimensions['x'] = self.listPos['x']
        newElement.dimensions['y'] = DynamicValue(self.getElementY, args={'index': i})

      self.elements.append(newElement)

  def getElementY(self, index: int) -> int | float:
    if index > 0:
      prevElement = self.elements[index-1]

      if hasattr(prevElement, 'section'):
        prevElementBase = prevElement.section.dimensions['y'].value + prevElement.section.dimensions['height'].value
      elif hasattr(prevElement, 'dimensions'):
        prevElementBase = prevElement.dimensions['y'].value + prevElement.dimensions['height'].value
      else:
        self.listPos['y'].resolveValue()

        return self.listPos['y'].value

      self.spacing.resolveValue()

      return prevElementBase + self.spacing.value

    self.listPos['y'].resolveValue()

    return self.listPos['y'].value



import pygame as pg
from pg_extended.Core.Base.DynamicValue import DynamicValue
from pg_extended.Core.Base.AnimatedValue import AnimatedValue

type NumValue = DynamicValue | AnimatedValue | int | float

class RectArea:
  def __init__(self, dimensions: dict[str, NumValue]):
    self.dimensions: dict[str, NumValue] = dimensions

    if not len(self.dimensions) == 4:
      raise ValueError(f'dimensions must contain 4 Dimension objects, received: {len(self.dimensions)}')

    for key in ('x', 'y', 'width', 'height'):
      if not key in self.dimensions:
        raise ValueError('dimensions must contain all of the following keys: \'x\', \'y\', \'width\' \'height\'')

    self.x: int | float
    self.y: int | float
    self.width: int | float
    self.height: int | float

    self.rect: pg.Rect = pg.Rect(0, 0, 0, 0)

    self.update()

  def getDimValue(self, key: str) -> int | float:
    return self.dimensions[key].value if isinstance(self.dimensions[key], (DynamicValue, AnimatedValue)) else self.dimensions[key]

  def update(self):
    for key in self.dimensions:
      if isinstance(self.dimensions[key], (DynamicValue, AnimatedValue)):
        self.dimensions[key].resolveValue()

    self.x = self.getDimValue("x")
    self.y = self.getDimValue("y")
    self.width = self.getDimValue("width")
    self.height = self.getDimValue("height")
    
    self.rect.update(self.x, self.y, self.width, self.height)



import pygame as pg

class Lifecycle:
  def openWindow(self):
    self.time = pg.time
    self.clock = self.time.Clock()
    self.currentFPS: int = self.clock.get_fps()

    pg.display.set_caption(self.title)

    self.screen = pg.display.set_mode((self.screenWidth, self.screenHeight), pg.RESIZABLE)

    receivedWidth, receivedHeight = self.screen.get_size()

    if ((not receivedWidth == self.screenWidth) or (not receivedHeight == self.screenHeight)):
      self.screenWidth, self.screenHeight = receivedWidth, receivedHeight

    self.running = True
    self.secondResize = False

    self.initiateActiveScene(self.screen)

    self.initiateActiveSystems(self.screen)

    if self.viewPort is not None:
      self.viewPort.initiate(self.screen, self.activeScene)

    self.resetUI()

    while self.running:
      self.updateLoop()

    self.closeWindow()

  def closeWindow(self):
    self.running = False
    self.deactivateSystems('all')
    self.deactivateScene()

    del self.screen
    del self.time
    del self.clock



from pg_extended.Types import CallableLike
import pg_extended as pgx

from .SystemManager import SystemManager
from .SceneManager import SceneManager
from .EventManager import EventManager
from .MainLoop import MainLoop
from .Lifecycle import Lifecycle
from .Utility import Utility

class Window(SystemManager, SceneManager, EventManager, MainLoop, Lifecycle, Utility):
  def __init__(self, title: str, screenRes: list[int] | tuple[int, int], customLoopProcess: CallableLike | None = None, customUpdateProcess: CallableLike | None = None, customEventHandler: CallableLike | None = None, customDrawProcess: CallableLike | None = None, fps : int | None = 60):
    self.title: str = title
    self.screenRes: list[int] | tuple[int, int] = screenRes
    self.customLoopProcess: CallableLike | None = customLoopProcess
    self.customEventHandler: CallableLike | None = customEventHandler
    self.customUpdateProcess: CallableLike | None = customUpdateProcess
    self.customDrawProcess: CallableLike | None = customDrawProcess
    self.screenWidth: int = self.screenRes[0]
    self.screenHeight: int = self.screenRes[1]
    self.fps: int = fps

    self.running: bool = False
    self.systems: dict[str, pgx.UI.System] = {}
    self.activeSystems: dict[str, pgx.UI.System] = {}
    self.systemZ: dict[str, int] = {}
    self.scenes: dict[str, pgx.Game.Scene] = {}
    self.activeScene: pgx.Game.Scene = None
    self.viewPort: pgx.Game.ViewPort = None
    self.customDynamicValues: dict[str, pgx.Core.DynamicValue] = {}
    self.lazyDynamicValues: dict[str, pgx.Core.DynamicValue] = {}
    self.customAnimatedValues: dict[str, pgx.Core.AnimatedValue] = {}
    self.customData: dict = {}
    self.firstUpdate = True



class Misc:
  # maps a number from one range to another
  @staticmethod
  def mapRange(num: float, start1: float, start2: float, end1: float, end2: float) -> float:
    # return the mid-point of the end range if the start1 and start2 are the same
    if start1 == start2:
      return (end1 + end2) / 2

    return end1 + (num - start1) * (end2 - end1) / (start2 - start1)

  # check if all the values are in an iterable if yes return True else return False
  @staticmethod
  def allIn(values: list | tuple, itr: list | tuple) -> bool:
    for v in values:
      if not v in itr:
        return False
    return True



from typing import Any
from pg_extended.Core.Base import DynamicValue, AnimatedValue
from pg_extended.Types import CallableLike

class Callback:
  def __init__(self, triggers: list[str] | tuple[str], func: CallableLike, staticArgs: dict[str, Any] = None, extraArgKeys: dict[str, str] = None):
    self.triggers = triggers
    self.func = func
    self.staticArgs = staticArgs or {}
    self.resolvedArgs = {}
    self.extraArgKeys = extraArgKeys or {}
    self.totalArgs = {}

  def _setExtraArgs(self, args: dict[str, Any] = None):
    args = args or {}

    self.totalArgs = self.staticArgs.copy()

    for key, value in args.items():
      if key in self.extraArgKeys:
        self.totalArgs[self.extraArgKeys[key]] = value

  def resolveArgs(self):
    self.resolvedArgs = {}

    for key in self.totalArgs:
      val = self.totalArgs[key]

      if isinstance(val, (DynamicValue, AnimatedValue)):
        val.resolveValue()
        self.resolvedArgs[key] = val.value
      else:
        self.resolvedArgs[key] = val

  def call(self, extraArgs: dict[str, Any] = None):
    extraArgs = extraArgs or {}

    self._setExtraArgs(extraArgs)

    self.resolveArgs()

    self.func(**self.resolvedArgs)

class CallbackSet:
  def __init__(self, callbacks: list[Callback] | tuple[Callback]):
    self.callbacks = callbacks
    self.callbacksDict = {}

    for callback in self.callbacks:
      for tgr in callback.triggers:
        self.callbacksDict.setdefault(tgr, []).append(callback)

  def call(self, trigger: str, extraArgs: dict[str, Any] = None):
    if trigger not in self.callbacksDict: return None

    for callback in self.callbacksDict[trigger]:
      callback.call(extraArgs)
